package lz.android.dexlib2.helper;

import org.jf.dexlib2.iface.ClassDef;
import org.jf.dexlib2.iface.Method;
import org.jf.dexlib2.immutable.ImmutableClassDef;

import java.util.ArrayList;
import java.util.List;

/**
* Enhancement of the ClassDef class operation.
* Created by buwai on 2015/4/1.
*/
public class ClassDefHelper {

    /**
* Determine the class whether there is a corresponding method.
* Judgment:
* Class definitions, method name, return type, parameters.
* @param classDef the class definition.
* @param method the method.
* @return true: exists.false: does not exist.
*/
    public static boolean isMethodExist(ClassDef classDef, Method method) {
        for (Method m : classDef.getMethods()) {
            // Determine the class name, method name.
            if (m.getDefiningClass().equals(method.getDefiningClass()) &&
                m.getName().equals(method.getName()) &&
                m.getReturnType().equals(method.getReturnType()) &&
                (m.getParameters().size() == method.getParameters().size())) {
                int size = m.getParameters().size();
                boolean bRet = true;
                for (int i = 0; i < size; i++) {
                    if (!m.getParameters().get(i).equals(method.getParameters().get(i))) {
                        bRet = false;
                        break;
                    }
                }
                return bRet;
            }
        }
        return false;
    }

    /**
* Determine the<clinit>method exists.
* @param classDef the class definition object.
* @return if it exists, then return the<clinit>method of the object.If it does not exist, null is returned.
*/
    public static Method isclinitMethodExit(ClassDef classDef) {
        for (Method m : classDef.getMethods()) {
            if ("<clinit>".equals(m.getName())) {
                return m;
            }
        }
        return null;
    }

    /**
* To ClassDef add a new method.
* @param classDef ClassDef object.
* @param method to insert the method.
* @return returns the new class definition.
*/
    public static ClassDef addMethod(ClassDef classDef, Method method) {
        List<Method> newMethods = new ArrayList<>();
        newMethods.add(method);
        for (Method m : classDef.getMethods()) {
            newMethods.add(m);
        }
        return new ImmutableClassDef(classDef.getType(), classDef.getAccessFlags(), classDef.getSuperclass(), classDef.getInterfaces(), classDef.getSourceFile(), classDef.getAnnotations(), classDef.getFields(), newMethods);
    }

    /**
* If the class does not exist in the corresponding method, then add.If there is, then replace.
* @param classDef the class definition.
* @param method the method.
* @return returns the new class definition.
*/
    public static ClassDef addOrReplaceMethod(ClassDef classDef, Method method) {
        if (!isMethodExist(classDef, method)) {
            return addMethod(classDef, method);
        }

        List<Method> newMethods = new ArrayList<>();
        for (Method m : classDef.getMethods()) {
            if (m.getDefiningClass().equals(method.getDefiningClass()) &&
                    m.getName().equals(method.getName()) &&
                    m.getReturnType().equals(method.getReturnType()) &&
                    (m.getParameters().size() == method.getParameters().size())) {
                int size = m.getParameters().size();
                boolean bRet = true;
                for (int i = 0; i < size; i++) {
                    if (!m.getParameters().get(i).getType().equals(method.getParameters().get(i).getType())) {
                        bRet = false;
                        break;
                    }
                }
                if (bRet) {
                    newMethods.add(method);
                } else {
                    newMethods.add(m);
                }
            } else {
                newMethods.add(m);
            }
        }
        return new ImmutableClassDef(classDef.getType(), classDef.getAccessFlags(), classDef.getSuperclass(), classDef.getInterfaces(), classDef.getSourceFile(), classDef.getAnnotations(), classDef.getFields(), newMethods);
    }

    /**
* Alternative method, do not do this method already exists in the judgment.
* So if the incoming method is not present, then the consequences are unknown.
* @param classDef the class definition.
* @param method to replace the method.
* @return returns the new class definition.
*/
    public static ClassDef replaceMethod(ClassDef classDef, Method method) {
        List<Method> newMethods = new ArrayList<>();
        for (Method m : classDef.getMethods()) {
            if (m.getDefiningClass().equals(method.getDefiningClass()) &&
                    m.getName().equals(method.getName()) &&
                    m.getReturnType().equals(method.getReturnType()) &&
                    (m.getParameters().size() == method.getParameters().size())) {
                int size = m.getParameters().size();
                boolean bRet = true;
                for (int i = 0; i < size; i++) {
                    if (!m.getParameters().get(i).getType().equals(method.getParameters().get(i).getType())) {
                        bRet = false;
                        break;
                    }
                }
                if (bRet) {
                    newMethods.add(method);
                } else {
                    newMethods.add(m);
                }
            } else {
                newMethods.add(m);
            }
        }
        return new ImmutableClassDef(classDef.getType(), classDef.getAccessFlags(), classDef.getSuperclass(), classDef.getInterfaces(), classDef.getSourceFile(), classDef.getAnnotations(), classDef.getFields(), newMethods);
    }

}

