package lz.android.shell.separator.config;

import lz.android.shell.base.Common;
import lz.android.shell.base.TypeDescription;
import lz.android.shell.base.helper.ResourceHelper;
import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.alibaba.fastjson.TypeReference;
import org.apache.log4j.Logger;
import org.jf.dexlib2.AccessFlags;
import org.jf.dexlib2.Opcodes;

import javax.annotation.Nullable;
import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.util.Map;

/**
* Created by buwai on 2015/4/1.
*/
public class ConfigParse {

private final static Logger log = Logger.getLogger(ConfigParse.class);

private File mConfigFile;

private Opcodes mOpcodes;

/**
* Constructor.
* @param configFile the configuration file.If you pass null, then parse the program stored in the default configuration file.
*/
    public ConfigParse(@Nullable File configFile) {
        mConfigFile = configFile;
        mOpcodes = new Opcodes(Common.API);
    }

    /**
* Parsing the configuration file.
*/
    public Config parse() {
        Config config = null;
        try {
            config = new Config();
            // Analysis system-property.conf file content.
            parseSystemProperty(config);

            // Parse the user's configuration file.
            parseUserConfig(config);

        } catch (IOException e) {
            config = null;
            e.printStackTrace();
        }
        return config;
    }

    /**
* Parsing system-property.conf file content.
* @param config the parsed data to the config assignment.
*/
    private void parseSystemProperty(Config config) throws IOException {
        // First read the system configuration properties.
        String systemProperty = ResourceHelper.getResourceTxt(this.getClass(), Config.CONF_FILE_SYSTEM_PROPERTY);
        log.info("systemProperty:" + systemProperty);
        Map<String, Object> sysConf = JSON.parseObject(systemProperty, new TypeReference<Map<String, Object>>() {
        });
        // Parse system configuration file in the blacklist.
        JSONArray blacklist = (JSONArray) sysConf.get(Config.CONF_BLACK_LIST);
        for (int i = 0; i < blacklist.size(); i++) {
            String item = blacklist.getString(i);
            TypeDescription td = parseTypeDescription(item);
            if (!config.blackList.contains(td)) {
                config.blackList.add(td);
            }
        }

        JSONObject no_supports = (JSONObject) sysConf.get(Config.CONF_NON_SUPPORTS);
        // Resolve does not support class access flags.
        JSONArray no_supports_class_acc_flags = (JSONArray) no_supports.get(Config.CONF_CLASS_ACC_FLAGS);
        for (int i = 0; i < no_supports_class_acc_flags.size(); i++) {
            String item = no_supports_class_acc_flags.getString(i);
            if (!config.nonsupportClassAccFlags.contains(item)) {
                config.nonsupportClassAccFlags.add(AccessFlags.getAccessFlag(item));
            }
        }
        // Resolve does not support the method access flags.
        JSONArray no_supports_method_acc_flags = (JSONArray) no_supports.get(Config.CONF_METHOD_ACC_FLAGS);
        for (int i = 0; i < no_supports_method_acc_flags.size(); i++) {
            String item = no_supports_method_acc_flags.getString(i);
            if (!config.nonsupportMethodAccFlags.contains(item)) {
                config.nonsupportMethodAccFlags.add(AccessFlags.getAccessFlag(item));
            }
        }
        // Resolve does not support the Opcode is.
        JSONArray no_supports_opcodes = (JSONArray) no_supports.get(Config.CONF_OPCODES);
        for (int i = 0; i < no_supports_opcodes.size(); i++) {
            String item = no_supports_opcodes.getString(i);
            if (!config.nonsupportOpcodes.contains(item)) {
                config.nonsupportOpcodes.add(mOpcodes.getOpcodeByName(item));
            }
        }

        // Whether to support the method in a try...catch block.
        String tryValue = no_supports.getString(Config.CONF_TRY);
        if (tryValue.equals("yes")) {
            config.isSupportTryCatch = true;
        } else {
            config.isSupportTryCatch = false;
        }
    }

    /**
* Parse the user's configuration file.
* @param config the parsed data to the config assignment.
*/
    private void parseUserConfig(Config config) throws IOException {
        String userProperty = null;
        // Then read the user's configuration file.
        if (null == mConfigFile) {   // If userConfigFile incoming null, then use the default configuration file.
            userProperty = ResourceHelper.getResourceTxt(this.getClass(), Config.CONF_FILE_USER_DEFAULT);
        } else {
            userProperty = new String(Files.readAllBytes(mConfigFile.toPath()));
        }

        Map<String, Object> userConf = JSON.parseObject(userProperty, new TypeReference<Map<String, Object>>() {
        });

        // Parse the blacklist.
        JSONArray blackList = (JSONArray) userConf.get(Config.CONF_BLACK_LIST);
        if (null != blackList) {
            for (int i = 0; i < blackList.size(); i++) {
                String item = blackList.getString(i);
                TypeDescription td = parseTypeDescription(item);
                if (!config.blackList.contains(td)) {
                    config.blackList.add(td);
                }
            }
        }

        // Parse the White list.
        JSONArray whiteList = (JSONArray) userConf.get(Config.CONF_WHITE_LIST);
        if (null != whiteList) {
            for (int i = 0; i < whiteList.size(); i++) {
                String item = whiteList.getString(i);
                TypeDescription td = parseTypeDescription(item);
                if (!config.whiteList.contains(td)) {
                    config.whiteList.add(td);
                }
            }
        }
    }

    /**
* Resolve the type described.
* @param item the black and white lists of one item.
* @return return the type description object.
*/
    private TypeDescription parseTypeDescription(String item) {
        String[] items = item.split("#");
        int len = items.length;
        TypeDescription td = new TypeDescription();
        if (3 == len) {
            String methodInfo = items[2];
            int methodSigStart = methodInfo.indexOf('(');
            if (-1 != methodSigStart) {
                int methodSigEnd = methodInfo.indexOf(')');
                if (-1 == methodSigEnd) {
                    log.error("Black and white list of item to write is a problem, the method signature write error:" + methodInfo);
                }
                td.methodName = methodInfo.substring(0, methodSigStart);
                td.methodSig = methodInfo.substring(methodSigStart);
            } else {
                td.methodName = methodInfo;
            }
        }
        if (len >= 2) {
            td.className = items[1];
        }
        td.packageName = items[0];
        return td;
    }
}

